import { DeviceConnectHandler, DeviceFuncCallHandler, DevicePropGetHandler, DevicePropSetHandler, DeviceThMdPostResHandler, devCallBack, mqttCallbackOnConnect } from "./message_callback"
import { EventPost, OnlineStatus } from "../protocol/payload/events"
import { FuncCallRes } from "../protocol/payload/functions"
import { PropGetRes, PropPost, PropSetRes } from "../protocol/payload/properties"
import { ThingModelPost } from "../protocol/payload/thingmodel"
import mqtt = require('mqtt')
var fs = require('fs');
import topic from "../protocol/topic/topic"
import util from "../util/util"
import { logger } from "../util/log"
export interface MqttConnOpt extends mqtt.IClientOptions { 
    gzip: boolean | undefined; // 是否开启gzip压缩
}

export interface JoylinkSDK {
    // 欢迎函数
    Welcome(): void

    // 初始化函数
    Initialise(): Error | void

    // 连接函数
    Connect(): Error | void

    // 设置设备回调 连接成功
    SetDevCallBackConnect(handler: DeviceConnectHandler): Error | void

    // 设备上下线状态上报
    OnlineStatusPost(objName: string, serviceKey: string, inPost: OnlineStatus): Error | void

    // 物模型上报函数
    ThingModelPost(objName: string, serviceKey: string, inPost: ThingModelPost): Error | void

    // 设置设备回调 物模型上报响应
    SetDevCallBackThMdPostRes(handler: DeviceThMdPostResHandler): Error | void

    // 属性上报
    PropertiesPost(objName: string, serviceKey: string, inPost: PropPost): Error | void

    // 设置设备回调 属性设置
    SetDevCallBackPropSet(handler: DevicePropSetHandler): Error | void

    // 属性设置响应
    PropertiesSetResponse(objName: string, serviceKey: string, inRes: PropSetRes): Error | void

    // 设置设备回调 属性获取
    SetDevCallBackPropGet(handler: DevicePropGetHandler): Error | void

    // 属性获取响应
    PropertiesGetResponse(objName: string, serviceKey: string, inRes: PropGetRes): Error | void

    // 设置设备回调 方法调用
    SetDevCallBackFuncCall(handler: DeviceFuncCallHandler): Error | void

    // 方法调用响应
    FunctionsCallResponse(objName: string, serviceKey: string, inRes: FuncCallRes): Error | void

    // 事件上报
    EventsPost(objName: string, serviceKey: string, inPost: EventPost): Error | void
}

// joylink 配置结构体定义
export type JoylinkConfig = {
    protocol?: string;
    hostname?: string;
    port?: number;
    deviceId?: string;
    username?: string;
    password?: string;
    cafile?: string;
    cert?: string;
    key?: string;
    insecure?: boolean;
    gzip?: boolean;
}

// sdk 结构体定义
type joylinkSDK = {
    mqtt: mqtt.MqttClient;
    version: string;
}

// 初始化sdk
function NewSDK(cfg: JoylinkConfig): JoylinkSDK {
    logger.info("[Joy SDk]. time=", Date.now())
    let opts: MqttConnOpt = {
        clientId: cfg.deviceId,
        username: cfg.username,
        password: cfg.password,
        gzip: cfg.gzip
    }
    let protocol = cfg.protocol
    let broker = protocol + "://" + cfg.hostname + ":" + String(cfg.port)
    if (protocol == "tls") {
        broker = "mqtts" + "://" + cfg.hostname + ":" + String(cfg.port)
        // 校验 ssl必填字段不能为空
        opts.ca = fs.readFileSync(cfg.cafile)
        opts.cert = fs.readFileSync(cfg.cert)
        opts.key = fs.readFileSync(cfg.key)
        opts.rejectUnauthorized = false;
    }
    logger.info(broker)
    logger.info("using gzip: ", opts.gzip)

    let s = new JoylinkSDKEntity(broker, "v1.0.0", opts)
    // opts.OnConnect = mqttCallbackOnConnect

    return s
}

// joylink实现类
export class JoylinkSDKEntity implements JoylinkSDK, joylinkSDK {
    mqtt!: mqtt.MqttClient
    version: string
    broker: string
    options: MqttConnOpt

    constructor(broker: string, version: string, options: MqttConnOpt) { // mqtt: mqtt.MqttClient
        // this.mqtt = mqtt.mqtt;
        this.version = version;
        this.broker = broker;
        this.options = options;
    }

    // 欢迎函数
    Welcome(): void | null {
        logger.info("welcome to use joylink sdk 5.0 in nodejs :%s\n\n", this.version)
    }

    // 初始化函数
    Initialise(): Error | void {
        logger.info("Enter JoylinkSDk Initialise\n")
        // set sdk handle to devCallBack
        devCallBack.sdk = this
        logger.info("Leave JoylinkSDk Initialise\n\n")
        return
    }

    // 连接函数
    Connect(): Error | void {
        this.mqtt = mqtt.connect(this.broker, this.options)
        let optReader = this.mqtt.options
        logger.info("Connectting to ", this.broker)

        this.mqtt.on("connect", () => {
            logger.info("Connectting to mqtt Successfuuly\n")
            mqttCallbackOnConnect(this.mqtt)
        });

        this.mqtt.on('error', (err: Error) => {
            logger.error("Connectting to mqtt Error.", err)
            throw new Error("Connectting to mqtt Error.")
        });
    }

    // 设置设备回调 连接成功
    SetDevCallBackConnect(handler: DeviceConnectHandler): Error | void {
        devCallBack.onConnect = handler;
        return
    }

    // 设备上下线状态上报
    OnlineStatusPost(objName: string, serviceKey: string, inPost: OnlineStatus): Error | void {
        logger.info("Enter JoylinkSDk OnlineStatusPost\n")
        if (!inPost.messageId || inPost.messageId.length == 0) {
            inPost.messageId = util.GetMessageId("m0")
        }
        if (!inPost.timestamp || inPost.timestamp == 0) {
            inPost.timestamp = util.GetTimestamp()
        }
        // 获取配置项
        let optReader = this.mqtt.options

        // Marshal json
        let msgPayload = JSON.stringify(inPost)

        // 判断消息主题
        let msgTopic: string
        if (inPost.events == "connected") {
            msgTopic = topic.Get(topic.EVT_ONLINE, objName, optReader.clientId, serviceKey)
        } else {
            msgTopic = topic.Get(topic.EVT_OFFLINE, objName, optReader.clientId, serviceKey)
        }
        this.mqtt.publish(msgTopic, util.DoGZip(msgPayload, this.options.gzip), { qos: 1, retain: false })
        messagePublishPrint(msgTopic, msgPayload)
        logger.info("Leave JoylinkSDk OnlineStatusPost\n\n")
        return
    }

    // 物模型上报函数
    ThingModelPost(objName: string, serviceKey: string, inPost: ThingModelPost): Error | void {
        logger.info("Enter JoylinkSDk ThingModelPost\n")
        if (!inPost.timestamp || inPost.timestamp == 0) {
            inPost.timestamp = util.GetTimestamp()
        }
        // 获取配置项
        let optReader = this.mqtt.options
        if (!inPost.messageId || inPost.messageId.length == 0) {
            inPost.messageId = util.GetMessageId("m1")
            // Marshal json
            let msgPayload = JSON.stringify(inPost)
            let msgTopic = topic.Get(topic.THINGMODEL_POST, objName, optReader.clientId, serviceKey)
            this.mqtt.publish(msgTopic, util.DoGZip(msgPayload, this.options.gzip), { qos: 1, retain: false })
            messagePublishPrint(msgTopic, msgPayload)
            logger.info("Leave JoylinkSDk ThingModelPost\n\n")
        }
        return
    }

    // 设置设备回调 物模型上报响应
    SetDevCallBackThMdPostRes(handler: DeviceThMdPostResHandler): Error | void {
        devCallBack.onThMdPostRes = handler
        return
    }

    // 属性上报
    PropertiesPost(objName: string, serviceKey: string, inPost: PropPost): Error | void {
        logger.info("Enter JoylinkSDk PropertiesPost\n")
        if (!inPost.messageId || inPost.messageId.length == 0) {
            inPost.messageId = util.GetMessageId("m2")
        }
        if (!inPost.timestamp || inPost.timestamp == 0) {
            inPost.timestamp = util.GetTimestamp()
        }
        // 获取配置项
        let optReader = this.mqtt.options

        // Marshal json
        let msgPayload = JSON.stringify(inPost)
        let msgTopic = topic.Get(topic.PROP_POST, objName, optReader.clientId, serviceKey)
        this.mqtt.publish(msgTopic, util.DoGZip(msgPayload, this.options.gzip), { qos: 1, retain: false })
        messagePublishPrint(msgTopic, msgPayload)
        logger.info("Leave JoylinkSDk PropertiesPost\n\n")
        return
    }

    // 设置设备回调 属性设置
    SetDevCallBackPropSet(handler: DevicePropSetHandler): Error | void {
        devCallBack.onPropSet = handler
        return
    }

    // 属性设置响应
    PropertiesSetResponse(objName: string, serviceKey: string, inRes: PropSetRes): Error | void {
        logger.info("Enter JoylinkSDk PropertiesSetResponse\n")
        if (!inRes.timestamp || inRes.timestamp == 0) {
            inRes.timestamp = util.GetTimestamp()
        }

        // 获取配置项
        let optReader = this.mqtt.options

        // Marshal json
        let msgPayload = JSON.stringify(inRes)
        let msgTopic = topic.Get(topic.PROP_SET_RES, objName, optReader.clientId, serviceKey)
        this.mqtt.publish(msgTopic, util.DoGZip(msgPayload, this.options.gzip), { qos: 1, retain: false })
        messagePublishPrint(msgTopic, msgPayload)
        logger.info("Leave JoylinkSDk PropertiesSetResponse\n\n")
        return
    }

    // 设置设备回调 属性获取
    SetDevCallBackPropGet(handler: DevicePropGetHandler): Error | void {
        devCallBack.onPropGet = handler
        return
    }

    // 属性获取响应
    PropertiesGetResponse(objName: string, serviceKey: string, inRes: PropGetRes): Error | void {
        logger.info("Enter JoylinkSDk PropertiesGetResponse\n")
        if (!inRes.timestamp || inRes.timestamp == 0) {
            inRes.timestamp = util.GetTimestamp()
        }

        // 获取配置项
        let optReader = this.mqtt.options

        // Marshal json
        let msgPayload = JSON.stringify(inRes)
        let msgTopic = topic.Get(topic.PROP_GET_RES, objName, optReader.clientId, serviceKey)
        this.mqtt.publish(msgTopic, util.DoGZip(msgPayload, this.options.gzip), { qos: 1, retain: false })
        messagePublishPrint(msgTopic, msgPayload)
        logger.info("Leave JoylinkSDk PropertiesGetResponse\n\n")
        return
    }

    // 设置设备回调 方法调用
    SetDevCallBackFuncCall(handler: DeviceFuncCallHandler): Error | void {
        logger.info("Enter JoylinkSDk SetDevCallBackFuncCall\n")
        devCallBack.onFuncCall = handler
        logger.info("Leave JoylinkSDk SetDevCallBackFuncCall\n\n")
        return
    }

    // 方法调用响应
    FunctionsCallResponse(objName: string, serviceKey: string, inRes: FuncCallRes): Error | void {
        logger.info("Enter JoylinkSDk FunctionsCallResponse\n")
        if (!inRes.timestamp || inRes.timestamp == 0) {
            inRes.timestamp = util.GetTimestamp()
        }

        // 获取配置项
        let optReader = this.mqtt.options

        // Marshal json
        let msgPayload = JSON.stringify(inRes)
        let msgTopic = topic.Get(topic.FUNC_CALL_RES, objName, optReader.clientId, serviceKey)
        this.mqtt.publish(msgTopic, util.DoGZip(msgPayload, this.options.gzip), { qos: 1, retain: false })
        messagePublishPrint(msgTopic, msgPayload)
        logger.info("Leave JoylinkSDk FunctionsCallResponse\n\n")
        return
    }

    // 事件上报
    EventsPost(objName: string, serviceKey: string, inPost: EventPost): Error | void {
        logger.info("Enter JoylinkSDk EventsPost\n")
        if (!inPost.messageId || inPost.messageId.length == 0) {
            inPost.messageId = util.GetMessageId("m3")
        }
        if (!inPost.timestamp || inPost.timestamp == 0) {
            inPost.timestamp = util.GetTimestamp()
        }

        // 获取配置项
        let optReader = this.mqtt.options

        // Marshal json
        let msgPayload = JSON.stringify(inPost)
        let msgTopic = topic.Get(topic.EVT_POST, objName, optReader.clientId, serviceKey)
        this.mqtt.publish(msgTopic, util.DoGZip(msgPayload, this.options.gzip), { qos: 1, retain: false })
        messagePublishPrint(msgTopic, msgPayload)
        logger.info("Leave JoylinkSDk EventsPost\n\n")
        return
    }

}

// 消息发送日志打印
function messagePublishPrint(t: string, p: string) {
    // logger.Pprintf("\\033[1;33;60mMessage Publish\r\n%s\r\n%s\\033[0m\r\n\r\n", t, p)
    logger.info("Message Publish",  "\r\n", t, "\r\n", p, "\r\n")
}

export default {
    NewSDK
}