import { JoylinkSDK } from "./joylink";
import { FuncCall } from "../protocol/payload/functions";
import { PropGet, PropSet } from "../protocol/payload/properties";
import { ThingModelPostRes } from "../protocol/payload/thingmodel";
import mqtt = require('mqtt')
import topic from "../protocol/topic/topic";
import { logger } from "../util/log";

// 设备回调函数句柄 连接成功
export type DeviceConnectHandler = (sdk: JoylinkSDK) => void;

// 设备回调函数句柄 物模型上报响应
export type DeviceThMdPostResHandler = (sdk: JoylinkSDK, objName: string, serviceKey: string, inRes: ThingModelPostRes) => void;

// 设备回调函数句柄 属性设置
export type DevicePropSetHandler = (sdk: JoylinkSDK, objName: string, serviceKey: string, inSet: PropSet) => void;

// 设备回调函数句柄 属性获取
export type DevicePropGetHandler = (sdk: JoylinkSDK, objName: string, serviceKey: string, inGet: PropGet) => void;

// 设备回调函数句柄 方法调用
export type DeviceFuncCallHandler = (sdk: JoylinkSDK, objName: string, serviceKey: string, inCall: FuncCall) => void;


// 设备回调函数句柄结构定义
export type deviceCallbackHandler = {
    sdk?: JoylinkSDK;              // JoylinkSDK
    onConnect?: DeviceConnectHandler;   // 连接成功 回调函数
    onThMdPostRes?: DeviceThMdPostResHandler;// 物模型上报响应 回调函数
    onPropSet?: DevicePropSetHandler;   // 属性设置 回调函数
    onPropGet?: DevicePropGetHandler;   // 属性获取 回调函数
    onFuncCall?: DeviceFuncCallHandler;  // 方法调用 回调函数
}


// 静态变量定义
// 设备回调函数
export var devCallBack: deviceCallbackHandler = {};

// MQTT 连接成功回调函数
export function mqttCallbackOnConnect(client: mqtt.MqttClient) {
    logger.info("Enter JoylinkSDk mqttCallbackOnConnect\n")

    // 获取配置项
    let optReader = client.options

    // 订阅消息
    var msgTopic: string[] = [
        topic.Get(topic.THINGMODEL_POST_RES, "+", optReader.clientId, ""), // 物模型上报响应
        topic.Get(topic.PROP_SET, "+", optReader.clientId, ""), // 通用属性设置
        topic.Get(topic.PROP_SET, "+", optReader.clientId, "+"), // 服务 属性设置
        topic.Get(topic.PROP_GET, "+", optReader.clientId, ""), // 通用属性获取
        topic.Get(topic.PROP_GET, "+", optReader.clientId, "+"), // 服务属性获取
        topic.Get(topic.FUNC_CALL, "+", optReader.clientId, ""), // 通用方法调用
        topic.Get(topic.FUNC_CALL, "+", optReader.clientId, "+"), // 服务方法调用
    ]

    // 循环遍历topic
    msgTopic.forEach(topic => {
        client.subscribe(topic, { qos: 1 }, (err, granted) => {
            err != null ? logger.error("subscribe topic error. ", granted, err) : logger.info("subscribe topic success:", granted)
        })
    })

    // 监听消息
    client.on("message", (topic: string, payload: Buffer) => {
        mqttCallbackOnMessage(client, topic, payload.toString())
    })

    // 设备回调
    if (devCallBack.onConnect && devCallBack.sdk) {
        devCallBack.onConnect(devCallBack.sdk)
    }

    logger.info("Leave JoylinkSDk mqttCallbackOnConnect\n\n")
}

// MQTT 消息处理回调函数
function mqttCallbackOnMessage(client: mqtt.MqttClient, msgTopic: string, payload: string) {
    logger.info("Enter JoylinkSDk mqttCallbackOnMessage\n")
    logger.info("joylink recv message\r\n%s\r\n%s\r\n ", msgTopic, payload)

    // 解析消息主题
    let parsedTopic = topic.Parse(msgTopic)

    // 消息主题分发流程
    if (parsedTopic.typeName == topic.TYPE_NAME_PROP) {
        // 属性消息 处理
        if (parsedTopic.command == "set") {
            // 属性设置
            onProcMessagePropertiesSet(msgTopic, payload, parsedTopic.objName, parsedTopic.serviceKey)
        } else if (parsedTopic.command == "get") {
            // 属性获取
            onProcMessagePropertiesGet(msgTopic, payload, parsedTopic.objName, parsedTopic.serviceKey)
        } else {

        }
    } else if (parsedTopic.typeName == topic.TYPE_NAME_THMD) {
        // 物模型 消息处理
        if (parsedTopic.command == "post" && parsedTopic.response == "response") {
            // 物模型上报 响应
            onProcMessageThingModelPostRes(msgTopic, payload, parsedTopic.objName, parsedTopic.serviceKey)
        }

    } else if (parsedTopic.typeName == topic.TYPE_NAME_FUNC) {
        // 方法调用 消息处理
        if (parsedTopic.command == "call") {
            // 属性设置
            onProcMessageFunctionsCall(msgTopic, payload, parsedTopic.objName, parsedTopic.serviceKey)
        }
    } else {
        // TODO 其他消息处理

    }

    logger.info("Leave JoylinkSDk mqttCallbackOnMessage\n\n")
}

// 物模型上报响应 消息处理函数
function onProcMessageThingModelPostRes(msgTopic: string, payload: string, objName: string, serviceKey: string) {
    logger.info("Enter JoylinkSDk onProcMessageThingModelPostRes\n")

    // 解析消息
    let postRes: ThingModelPostRes = JSON.parse(payload)

    if (devCallBack.onThMdPostRes && devCallBack.sdk) {
        devCallBack.onThMdPostRes(devCallBack.sdk, objName, serviceKey, postRes)
    }
    logger.info("Leave JoylinkSDk onProcMessageThingModelPostRes\n\n")
}

// 属性设置 消息处理函数
function onProcMessagePropertiesSet(msgTopic: string, payload: string, objName: string, serviceKey: string) {
    logger.info("Enter JoylinkSDk onProcMessagePropertiesSet\n")

    // 解析消息
    let propSet: PropSet = JSON.parse(payload)

    if (devCallBack.onPropSet && devCallBack.sdk) {
        devCallBack.onPropSet(devCallBack.sdk, objName, serviceKey, propSet)
    }
    logger.info("Leave JoylinkSDk onProcMessagePropertiesSet\n\n")
}

// 属性获取 消息处理函数
function onProcMessagePropertiesGet(msgTopic: string, payload: string, objName: string, serviceKey: string) {
    logger.info("Enter JoylinkSDk onProcMessagePropertiesGet\n")

    // 解析消息
    let propGet: PropGet = JSON.parse(payload)

    if (devCallBack.onPropGet && devCallBack.sdk) {
        devCallBack.onPropGet(devCallBack.sdk, objName, serviceKey, propGet)
    }
    logger.info("Leave JoylinkSDk onProcMessagePropertiesGet\n\n")
}

// 方法调用 消息处理函数
function onProcMessageFunctionsCall(msgTopic: string, payload: string, objName: string, serviceKey: string) {
    logger.info("Enter JoylinkSDk onProcMessageFunctionsCall\n")

    // 解析消息
    let funcCall: FuncCall = JSON.parse(payload)

    if (devCallBack.onFuncCall && devCallBack.sdk) {
        devCallBack.onFuncCall(devCallBack.sdk, objName, serviceKey, funcCall)
    }
    logger.info("Leave JoylinkSDk onProcMessageFunctionsCall\n\n")
}

