import { JoylinkSDK } from "../../core/joylink"
import { FuncCall } from "../../protocol/payload/functions"
import { PropGet, PropSet } from "../../protocol/payload/properties"
import { ThingModel, ThingModelPostRes } from "../../protocol/payload/thingmodel"
import { logger } from "../../util/log"
import entity from "./entity"

/*
 * 方法调用相关参数定义 仅为测试使用 真实设备时请根据真实情况实现
 */
const DEV_FUN_TYPE_NONE: number = 0 // 无入参
const DEV_FUN_TYPE_INOUT: number = 1 // 直接返回入参

type DevFunCall = {
	key: string;
	type: number;
	outNum: number;
	out: { [key: string]: any };
	outKey: string[];
}

// 定义设备结构体
type Device = {
	sdk: JoylinkSDK;    // sdk 句柄
	deviceId: string;               // 设备标识
	thingModel?: ThingModel;     // 物模型
	version: number;               // 属性版本
	propNum?: number;                   // 属性个数
	prop?: { [key: string]: any }; // 设备属性
	funNum?: number;                    //方法个数
	func?: DevFunCall[];       // 方法调用
}

class DeviceEntity implements Device {
	sdk: JoylinkSDK
	deviceId: string
	thingModel?: ThingModel
	version: number
	propNum?: number
	prop?: { [key: string]: any }
	funNum?: number
	func?: DevFunCall[]

	// 设备初始化
	constructor(deviceId: string, sdk: JoylinkSDK) {
		this.sdk = sdk
		this.deviceId = deviceId

		// 物模型
		this.thingModel = {
			id: "urn:user-spec-v1:thing-model:EC-control-test:bbd1cfc0279091635311806bcbc348f1",
			"thing-model-version": "v1.0.0"
		}

		/***************************************************************************
		 * 通用 属性
		 ***************************************************************************/
		this.version = 0
		this.propNum = 2
		this.prop = {
			"selfmode.propA": true,
			"selfmode.propB": 35
		}

		/***************************************************************************
		 * 通用 方法调用
		 ***************************************************************************/
		this.funNum = 1
		this.func = [{
			key: "selfmode.funcA",
			type: DEV_FUN_TYPE_INOUT,
			outNum: 1,
			out: {
				"out1": 20
			},
			outKey: ["out1"]
		}]
	}

}

// 定义设备
export var deviceEntity: Device;

// 设备初始化函数
function Initialise(deviceId: string | undefined, sdk: JoylinkSDK) {
	if (deviceId) {
		deviceEntity = new DeviceEntity(deviceId, sdk)
	} else {
		logger.error("deviceId is undefined. cannot initialise.")
	}
}

// 设备回调 SDK连接成功回调函数
function CallBackOnConnect(s: JoylinkSDK) {
	logger.info("\nEnter Device CallBackOnConnect")
	// 设备上线
	entity.entityOnlineStatusPost(true)

	// 物模型上传
	entity.entityThingModelPost()

	// 通用属性上报
	entity.entityPropertiesPost()

	// 通用事件上报
	entity.entityEventPost()

	logger.info("Leave Device CallBackOnConnect")
}

// 设备回调 物模型上报响应
function CallBackOnThingModelPostRes(s: JoylinkSDK, objName: string, serviceKey: string, inRes: ThingModelPostRes) {
	logger.info("Enter Device CallBackOnThingModelPostRes")
	logger.info("\tDeviceId:\t", inRes.deviceId)
	logger.info("\tMessageId:\t", inRes.messageId)
	logger.info("\tTimestamp:\t", inRes.timestamp)
	logger.info("\tCode:\t", inRes.code)
	logger.info("\tMessage:\t", inRes.message)
	logger.info("Leave Device CallBackOnThingModelPostRes")
}

// 设备回调 属性设置
function CallBackOnPropertiesSet(s: JoylinkSDK, objName: string, serviceKey: string, inSet: PropSet) {
	logger.info("Enter Device CallBackOnPropertiesSet")
	if (serviceKey == "") {
		// 通用属性设置
		entity.entityPropertiesSet(inSet)

	} else if (serviceKey == "connection_agent") {
		// connection agent 属性设置
		logger.info("connection agent properties set unsupported yet!!")

	} else {
	}
	logger.info("Leave Device CallBackOnPropertiesSet")
}

// 设备回调 属性获取
function CallBackOnPropertiesGet(s: JoylinkSDK, objName: string, serviceKey: string, inGet: PropGet) {
	logger.info("Enter Device CallBackOnPropertiesGet")
	if (serviceKey == "") {
		// 通用属性设置
		entity.entityPropertiesGet(inGet)

	} else if (serviceKey == "connection_agent") {
		// connection agent 属性设置
		logger.info("connection agent properties set unsupported yet!!")

	} else {
	}
	logger.info("Leave Device CallBackOnPropertiesGet")
}

// 设备回调 方法调用
function CallBackOnFunctionsCall(s: JoylinkSDK, objName: string, serviceKey: string, inCall: FuncCall) {
	logger.info("Enter Device CallBackOnFunctionsCall")
	if (serviceKey == "") {
		// 通用方法调用
		entity.entityFunctionsCall(inCall)

	} else if (serviceKey == "connection_agent") {
		// connection agent 属性设置
		logger.info("connection agent properties set unsupported yet!!")

	} else {
	}
	logger.info("Leave Device CallBackOnFunctionsCall")
}


export default {
	CallBackOnConnect,
	CallBackOnThingModelPostRes,
	CallBackOnPropertiesSet,
	CallBackOnPropertiesGet,
	CallBackOnFunctionsCall,
	Initialise
}